extern crate flfire;
extern crate rand;
use rand::prelude::*;

use std::time::Duration;
use std::cmp::{min,max};

const WIDTH: usize = 16;
const HEIGHT: usize = 4;
const TO: Duration = Duration::from_millis(50);

const RANDOM_FLICKER: bool = true;

type Grid = [[bool; WIDTH]; HEIGHT];

fn main() {
    let blocks: Vec<Box<dyn Block>> = vec![
        Box::new(Line()),
        Box::new(TBlock()),
        Box::new(SBlock()),
        Box::new(ZBlock()),
        Box::new(Square()),
    ];

    let cont = libusb::Context::new().expect("Failed to initialize a libusb context");
    let fires = flfire::get_fires(&cont);
    
    println!("{} FIRE's detected.", fires.len());
    assert_eq!(fires.len() / (HEIGHT / 4) * 16, WIDTH, "Incorrect amount of FIRE's");

    
    let mut handles = flfire::claim_fires(&fires);

    // Clear all LEDS
    handles.iter_mut().for_each(|h| { let _ = h.write_bulk(1, &[0x0b,0xb0,0x7f,0], TO).expect("Failed to clear LEDS"); });

    // Order FIRE's by hand
    // Fetch input from all devices
    // and insert the handles in the order of button presses
    let mut handles_ord = Vec::with_capacity(fires.len());
    let mut initial_bounce: u8 = 5;
    println!("Select device order from bottom left to top right by pressing a button.");
    loop {
        let mut received = Vec::with_capacity(fires.len());
        for (i, buf) in flfire::get_input(&handles, TO).iter().enumerate() {
            for o in 0..5 {
                if initial_bounce == 0 && buf[o*4] == 0x09 && !received.contains(&i) {
                    println!("FIRE {} pressed!", i);
                    received.push(i);
                }
            }
        }
        if received.len() > 1 {
            eprintln!("Received input from more than one FIRE, please reselect.");
            received.clear();
        } else if received.len() == 1 {
            let pos = received.pop().unwrap();
            let _ = handles[pos].write_bulk(0x1, &[0x0b,0xb0,0x33,0x02], TO).unwrap();
            handles_ord.push(handles.remove(pos));
        }
        if handles.len() == 0 {
            break;
        }
        initial_bounce = initial_bounce.saturating_sub(1);
    }

    let mut handles = handles_ord;

    'replayloop: loop {
        let mut grid = [[false; WIDTH]; HEIGHT];
        
        let mut fall_y: usize = HEIGHT / 2;
        let mut fall_x: usize = WIDTH - 1;
        let mut fall_rot = Rotation::Up;
        let mut fall_block: usize = random::<usize>() % blocks.len();

        let mut last_clear = None;
        let mut score: usize = 0;
        
        let mut activity_led = 2;
        let mut speed: usize = 10;
        let mut fall_timeout: usize = 0;

        for y in grid.iter_mut() {
            for (i, x) in y.iter_mut().enumerate() {
                if i < 5 {
                    *x = random();
                }
            }
        }

        /*
        FIRE0 FIRE2 .
        FIRE1 FIRE3 .
        .     .     .
        .     .     .
         */

        'main: loop {
            // Draw grid
            let mut y_active = [false; HEIGHT];
            for y in 0..HEIGHT {
                for x in 0..WIDTH {
                    let pad_buf: [u8; 16] = [
                        0x04, 0xf0, 0x47, 0x7f,
                        0x04, 0x43, 0x65, 0x00,
                        0x04, 0x04, (y % 4) as u8 *16 + (x % 16) as u8, if last_clear == Some(x) {
                            0x20
                        } else {
                            0x00
                        },
                        0x07, if blocks[fall_block].render(fall_x, fall_y, fall_rot,
                                                           x, y) {
                            // This is the position of the falling block
                            y_active[y] = true;
                            if fall_x == x && fall_y == y {
                                0x20
                            } else {
                                if RANDOM_FLICKER {
                                    random::<u8>() % 0x50 + 0x2f
                                } else {
                                    0x50
                                }
                            }
                        } else {
                            0x00
                        }, if grid[y as usize][x as usize] {
                            // There is a static block here
                            if RANDOM_FLICKER {
                                random::<u8>() % 0x20 + 0x4f
                            } else {
                                0x50
                            }
                        } else {
                            0x00
                        }, 0xf7
                    ];
                    handles[(y / 4) + (x / 16) * (HEIGHT / 4)].write_bulk(1, &pad_buf, TO).unwrap();
                }
            }
            // Show y position
            for (i, h) in handles.iter_mut().enumerate() {
                const LEDS: [u8; 4] = [0x28, 0x29, 0x2a, 0x2b];
                for (i_p, p) in LEDS.iter().enumerate() {
                    if y_active[i % (HEIGHT / 4) * 4 + i_p] {
                        let _ = h.write_bulk(1, &[0x0b,0xb0,*p,0x02], TO).unwrap();
                    } else {
                        let _ = h.write_bulk(1, &[0x0b,0xb0,*p,0x00], TO).unwrap();
                    }
                }
            }

            let mut buf = [0u8; 5*4];
            for b in flfire::get_input(&handles, TO) {
                buf = b;
                if buf.iter().find(|v| **v != 0) != None {
                    break;
                }
            }

            for (fire_i, o) in (0..5).enumerate() {
                if buf[o*4] == 0x0b /*&& buf[o*4+2] == 0x10*/ {
                    match buf[o*4+3] {
                        0x01 => if blocks[fall_block].valid_move(&grid, fall_x, fall_y,
                                                                 fall_rot, Direction::Right) {
                            fall_y = min(HEIGHT - 1, fall_y + 1);
                        },
                        0x7f => if blocks[fall_block].valid_move(&grid, fall_x, fall_y,
                                                                 fall_rot, Direction::Left) {
                            fall_y = fall_y.saturating_sub(1);
                        },
                        _ => {},
                    }
                    break;
                }
                if buf[o*4] == 0x09 {
                    match buf[o*4+2] {
                        0x2c | 0x19 => {
                            // Quick drop
                            while fall_x != 0 && !blocks[fall_block].collides(&grid, fall_x, fall_y, fall_rot) {
                                fall_x -= 1;
                                fall_timeout = speed;
                            }
                        },
                        0x2d => {
                            // Rotate clockwise
                            let old_rot = fall_rot;
                            fall_rot =  match fall_rot {
                                Rotation::Up => Rotation::Right,
                                Rotation::Right => Rotation::Down,
                                Rotation::Down => Rotation::Left,
                                Rotation::Left => Rotation::Up,
                            };
                            if !blocks[fall_block].valid_rotate(&grid, fall_x, fall_y, fall_rot) {
                                fall_rot = old_rot;
                            }
                        },
                        0x2e => {
                            // Rotate counter clockwise
                            let old_rot = fall_rot; // DUP
                            fall_rot =  match fall_rot {
                                Rotation::Right => Rotation::Up,
                                Rotation::Down => Rotation::Right,
                                Rotation::Left => Rotation::Down,
                                Rotation::Up => Rotation::Left,
                            };
                            if !blocks[fall_block].valid_rotate(&grid, fall_x, fall_y, fall_rot) {
                                fall_rot = old_rot;
                            }
                        },
                        0x34 => break 'main, // End game
                        pad_y if pad_y >= 0x36 && pad_y < 0x66 + 16 => {
                            // Quick drop at y
                            fall_y = ((pad_y as usize) - 0x36) / 16 * (fire_i % (HEIGHT / 4)) + 1;
                            
                            while fall_x != 0 && !blocks[fall_block].collides(&grid, fall_x, fall_y, fall_rot) {
                                fall_x -= 1;
                                fall_timeout = speed;
                            }
                        }
                        _ => {},
                    }
                }
            }

            if fall_timeout >= speed {
                fall_timeout = 0;
                // Block fall
                if fall_x == 0 || blocks[fall_block].collides(&grid, fall_x, fall_y, fall_rot) {
                    for x in 0..WIDTH {
                        for y in 0..HEIGHT {
                            // Put block into grid
                            grid[y][x] = grid[y][x] || blocks[fall_block].render(fall_x, fall_y, fall_rot,
                                                                                 x, y);
                        }
                    }
                    fall_block = random::<usize>() % blocks.len();
                    fall_x = WIDTH;
                    let new = blocks[fall_block].limit_entry(fall_x,fall_y,fall_rot);
                    fall_x = new.0;
                    fall_y = new.1;
                }
                fall_x = fall_x.saturating_sub(1);
                
                if grid[fall_y][fall_x] {
                    break 'main;
                }
                score = score.saturating_add(1);

                let chained = last_clear.is_some();
                last_clear = update_grid(&mut grid);
                if last_clear.is_some() {
                    score = score.saturating_add(4 * if chained { 2 } else { 1 });
                }
                activity_led ^= 2;
                handles.iter_mut().for_each(|h| { let _ = h.write_bulk(0x1, &[0x0b,0xb0,0x33,activity_led], TO).unwrap(); });
            }
            fall_timeout += 1;
        }

        println!("Game over!");
        println!("Score: {}", score);

        let mut first_clear = true;
        'endcheck: loop {
            for x in 0..WIDTH {
                for y in 0..HEIGHT {
                    let x = WIDTH - 1 - x;
                    let pad_buf: [u8; 16] = [
                        0x04, 0xf0, 0x47, 0x7f,
                        0x04, 0x43, 0x65, 0x00,
                        0x04, 0x04, (y % 4) as u8 *16 + (x % 16) as u8, if RANDOM_FLICKER {
                            random::<u8>() % 0x40 + 0x10
                        } else {
                            0x20
                        },
                        0x07, 0x00, 0x00, 0xf7
                    ];
                    handles[(y / 4) + (x / 16) * (HEIGHT / 4)].write_bulk(1, &pad_buf, TO).unwrap();
                }

                if first_clear {
                    std::thread::sleep_ms(40);
                }
            }
            if first_clear {
                first_clear = false;
            }

            let mut buf = [0u8; 5*4];
            for i in 0..handles.len() {
                let _ = handles[i].read_bulk(0x81, &mut buf, TO);
                if buf.iter().find(|v| **v != 0) != None {
                    break;
                }
            }
            
            for o in 0..5 {
                if buf[o*4] == 0x09 {
                    match buf[o*4+2] {
                        0x34 => { break 'replayloop; },
                        _ => {},
                    }
                    break 'endcheck;
                }
            }
        }
    }
}

fn update_grid(grid: &mut Grid) -> Option<usize> {
    // Currently this clears lines one at a time,
    // This looks kinda good but isn't very Tetris'y
    let mut chain = false;
    let mut pos = 0;
    for x in 0..WIDTH {
        let mut complete = true;
        for y in 0..HEIGHT {
            complete = complete && grid[y][x];
        }
        if complete {
            chain = true;
            pos = x;
        }
        if chain {
            // Move lines downward
            for y in 0..HEIGHT {
                grid[y][x] = if x < WIDTH - 1 && (!grid[y][x] || complete) && grid[y][x+1] {
                    grid[y][x+1] = false;
                    true
                } else {
                    false
                };
            }
        }
    }
    if chain {
        Some(pos)
    } else {
        None
    }
}

#[derive(Copy, Clone, Debug)]
enum Rotation {
    Up, Down, Right, Left,
}

enum Direction {
    Left, Right,
}

// NOTE All base implementations assume that the Block is never outside of the grid and continuous
trait Block {
    fn render(&self, x: usize, y: usize, rot: Rotation, r_x: usize, r_y: usize) -> bool; // FIXME Inconsistent interfaces
    
    fn collides(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        if x == 0 {
            return true
        }
        for p_x in 1..WIDTH {
            for p_y in 0..HEIGHT {
                if self.render(x,y,rot,p_x,p_y) && grid[p_y][p_x - 1] {
                    return true;
                }
            }
        }
        false
    }
    
    fn valid_move(&self, grid: &Grid, x: usize, y: usize, rot: Rotation, dir: Direction) -> bool {
        for p_x in 0..WIDTH {
            for p_y in 0..HEIGHT {
                match dir {
                    Direction::Left => {
                        if self.render(x,y,rot,p_x,p_y) && (p_y == 0 || grid[p_y - 1][p_x]) {
                            return false;
                        }
                    },
                    Direction::Right => {
                        if self.render(x,y,rot,p_x,p_y) && (p_y == HEIGHT-1 || grid[p_y + 1][p_x]) {
                            return false;
                        }
                    }
                }
            }
        }
        true
    }

    fn valid_rotate(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        let mut blocks_in_grid = 0;
        for p_x in 0..WIDTH {
            for p_y in 0..HEIGHT {
                if self.render(x, y, rot, p_x, p_y) {
                    blocks_in_grid += 1;
                    if grid[p_y][p_x] {
                        return false;
                    }
                }
            }
        }
        if blocks_in_grid != self.blocks_in_grid(rot) {
            false
        } else {
            true
        }
    }

    fn blocks_in_grid(&self, rot: Rotation) -> usize;
    
    fn limit_entry(&self, x: usize, y: usize, rot: Rotation) -> (usize, usize);
}

struct Line();
impl Block for Line {
    fn collides(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        match rot {
            Rotation::Up | Rotation::Down => for o_y in 0..2 {
                let act_y = (y + o_y).saturating_sub(1);
                if x == 0 || grid[act_y][x-1] {
                    return true
                }
            },
            Rotation::Right | Rotation::Left => if x == 1 || grid[y][x - 2] {
                return true
            },
        }
        false
    }

    fn render(&self, x: usize, y: usize, rot: Rotation, r_x: usize, r_y: usize) -> bool {
        match rot {
            Rotation::Up | Rotation::Down => if r_x == x && y > 0 && r_y >= y - 1 && r_y <= y + 1 {
                return true
            },
            Rotation::Right | Rotation::Left => if r_y == y && x > 0 && r_x >= x - 1 && r_x <= x + 1 {
                return true
            },
        }
        false
    }

    fn valid_move(&self, grid: &Grid, x: usize, y: usize, rot: Rotation, dir: Direction) -> bool {
        match rot {
            Rotation::Up | Rotation::Down => match dir {
                Direction::Left => y > 1,
                Direction::Right => y < HEIGHT - 2,
            },
            Rotation::Right | Rotation::Left => match dir {
                Direction::Left => y > 0,
                Direction::Right => y < HEIGHT - 1,
            },
        }
    }

    fn valid_rotate(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        match rot {
            Rotation::Up | Rotation::Down => y > 0 && y < HEIGHT - 1,
            Rotation::Left | Rotation::Right => true,
        }
    }

    fn blocks_in_grid(&self, _: Rotation) -> usize {
        4
    }

    fn limit_entry(&self, x: usize, y: usize, rot: Rotation) -> (usize, usize) {
        match rot {
            Rotation::Up => (x, max(1, y)),
            Rotation::Down => (x, min(HEIGHT-2, y)),
            Rotation::Left | Rotation::Right => (x, y),
        }
    }
}

struct TBlock();
impl Block for TBlock {
    fn render(&self, x: usize, y: usize, rot: Rotation, r_x: usize, r_y: usize) -> bool {
        match rot {
            Rotation::Up => if (r_x == x && (r_y as isize - y as isize).abs() <= 1)
                || (r_x == x + 1 && r_y == y) {
                return true
            },
            Rotation::Down => if (r_x == x + 1 && (r_y as isize - y as isize).abs() <= 1)
                || (r_x == x && r_y == y) {
                return true
            },
            Rotation::Right => if (r_y == y && (r_x as isize - x as isize).abs() <= 1)
                || (r_y == y + 1 && r_x == x) {
                return true
            },
            Rotation::Left => if (r_y == y && (r_x as isize - x as isize).abs() <= 1)
                || (r_y + 1 == y && r_x == x) {
                return true
            },
        }
        false
    }

    /*
    fn valid_rotate(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        y > 0 && y < HEIGHT - 1
    }
     */

    fn blocks_in_grid(&self, _: Rotation) -> usize {
        4
    }

    fn limit_entry(&self, x: usize, y: usize, rot: Rotation) -> (usize, usize) {
        match rot {
            Rotation::Up => (x, max(1, y)),
            Rotation::Down => (x, min(HEIGHT-2, y)),
            Rotation::Right | Rotation::Left => (x, max(1, min(HEIGHT-2, y))),
        }
    }
}

struct SBlock();
impl Block for SBlock {
    fn render(&self, x: usize, y: usize, rot: Rotation, r_x: usize, r_y: usize) -> bool {
        match rot {
            Rotation::Up | Rotation::Down => {
                if (r_y == y && (r_x == x || r_x == x + 1))
                    || (r_y == y + 1 && (r_x == x || r_x + 1 == x)) {
                    return true
                }
            },
            Rotation::Right | Rotation::Left => {
                if (r_x == x + 1 && (r_y == y || r_y == y + 1))
                    || (r_x == x && (r_y == y || r_y + 1 == y)) {
                    return true
                }
            },
        }
        false
    }

    /*
    fn valid_rotate(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        y > 0 && y < HEIGHT - 1
    }
     */

    fn blocks_in_grid(&self, _: Rotation) -> usize {
        4
    }
    
    fn limit_entry(&self, x: usize, y: usize, rot: Rotation) -> (usize, usize) {
        match rot {
            Rotation::Up | Rotation::Down => (x, y),
            Rotation::Right | Rotation::Left => (x, max(1, min(HEIGHT-2, y))),
        }
    }
}

struct ZBlock();
impl Block for ZBlock {
    fn render(&self, x: usize, y: usize, rot: Rotation, r_x: usize, r_y: usize) -> bool {
        match rot {
            Rotation::Up | Rotation::Down => {
                if (r_y == y + 1 && (r_x == x || r_x == x + 1))
                    || (r_y == y && (r_x == x || r_x + 1 == x)) {
                    return true
                }
            },
            Rotation::Right | Rotation::Left => {
                if (r_x == x && (r_y == y || r_y == y + 1))
                    || (r_x == x + 1 && (r_y == y || r_y + 1 == y)) {
                    return true
                }
            },
        }
        false
    }

    /*
    fn valid_rotate(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        y > 0 && y < HEIGHT - 1
    }
    */

    fn blocks_in_grid(&self, _: Rotation) -> usize {
        4
    }
    
    fn limit_entry(&self, x: usize, y: usize, rot: Rotation) -> (usize, usize) {
        match rot {
            Rotation::Up | Rotation::Down => (x, y),
            Rotation::Right | Rotation::Left => (x, min(1, max(HEIGHT-2, y))),
        }
    }
}

struct Square();
impl Block for Square {
    fn render(&self, x: usize, y: usize, rot: Rotation, r_x: usize, r_y: usize) -> bool {
        (r_x == x || r_x == x + 1) && (r_y == y || r_y == y + 1)
    }

    fn valid_rotate(&self, grid: &Grid, x: usize, y: usize, rot: Rotation) -> bool {
        true
    }

    fn blocks_in_grid(&self, _: Rotation) -> usize {
        4
    }

    fn limit_entry(&self, x: usize, y: usize, rot: Rotation) -> (usize, usize) {
        (x, min(y, HEIGHT - 2))
    }
}
